SystemOrganization addCategory: #'NativeBoost-Mac'!

SharedPool subclass: #NBMacConstants
	instanceVariableNames: ''
	classVariableNames: 'MAP_ANON MAP_COPY MAP_FAILED MAP_FILE MAP_FIXED MAP_HASSEMAPHORE MAP_NOCACHE MAP_NOEXTEND MAP_NORESERVE MAP_PRIVATE MAP_RENAME MAP_RESERVED0080 MAP_SHARED PROT_EXEC PROT_NONE PROT_READ PROT_WRITE RTLD_DEFAULT RTLD_FIRST RTLD_GLOBAL RTLD_LAZY RTLD_LOCAL RTLD_MAIN_ONLY RTLD_NEXT RTLD_NODELETE RTLD_NOLOAD RTLD_NOW RTLD_SELF'
	poolDictionaries: ''
	category: 'NativeBoost-Mac'!

!NBMacConstants class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/5/2010 19:18'!
initMmapflags

	PROT_READ := 16r1.	" Page can be read.  "
	PROT_WRITE := 16r2.	" Page can be written. "
	PROT_EXEC := 16r4.	" Page can be executed. "
	PROT_NONE := 16r0. 	" Page can not be accessed. "

	MAP_FILE := 0. " map from file (default) "
	MAP_ANON := 16r1000. " allocated from memory, swap space "
	
	" Flags contain sharing type and options.
 	Sharing types; choose one."
	
	MAP_SHARED := 16r0001. " [MF|SHM] share changes"
	MAP_PRIVATE := 16r0002. " [MF|SHM] changes are private"
	MAP_COPY  := MAP_PRIVATE.

	MAP_FIXED := 16r0010. " [MF|SHM] interpret addr exactly "
	
	MAP_RENAME := 16r0020. " Sun: rename private pages to file "
	MAP_NORESERVE := 16r0040. " Sun: don't reserve needed swap area "
	MAP_RESERVED0080 := 16r0080. " previously unimplemented MAP_INHERIT "
	MAP_NOEXTEND := 16r0100. " for MAP_FILE, don't change file size "
	MAP_HASSEMAPHORE := 16r0200. " region may contain semaphores "
	MAP_NOCACHE := 16r0400. " don't cache pages for this mapping "

	MAP_FAILED := -1 asImm asDWord.! !

!NBMacConstants class methodsFor: 'as yet unclassified' stamp: 'IgorStasenko 11/18/2012 17:08'!
initialize

	RTLD_LAZY := 16r1.
	RTLD_NOW := 16r2.
	RTLD_LOCAL := 16r4.
	RTLD_GLOBAL := 16r8.
	
	RTLD_NOLOAD := 16r10.
	RTLD_NODELETE := 16r80.
	RTLD_FIRST := 16r100.  " Mac OS X 10.5 and later "
	

	RTLD_NEXT := -1 . " ((void *) -1) Search subsequent objects. "
	RTLD_DEFAULT := -2. " ((void *) -2) Use default search algorithm. "
	RTLD_SELF := -3. " ((void *) -3)  Search this and subsequent objects (Mac OS X 10.5 and later) "
	RTLD_MAIN_ONLY := -5. " ((void *) -5)  Search main executable only (Mac OS X 10.5 and later)"

	self initMmapflags.
	! !

NBExternalHeapManager subclass: #NBMacExternalHeapManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'NBMacConstants'
	category: 'NativeBoost-Mac'!

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/5/2010 19:13'!
mapFlags

	^ MAP_ANON bitOr: MAP_PRIVATE! !

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'IgorStasenko 11/24/2012 17:22'!
mmapLength: bytesToAllocate prot: protFlags flags: mapFlags into: returnValueBuffer
		
	<primitive: #primitiveNativeCall module: #NativeBoostPlugin error: errorCode>

	"Note:  mmap using 32-bit offset, while mmap64 - 64bit one"

	^ self nbCallout 
		options: #( - optDirectProxyFnAddress );
		function: #(
			NBBootstrapUlong ( 0, uint bytesToAllocate, int protFlags, int mapFlags,
				-1,  0 "offset" )) 
		emit:  [:gen | | mmap |
			mmap := NativeBoost forCurrentPlatform getGlobalSymbolPointer: 'mmap'.
			gen asm mov: mmap asUImm32 to: gen asm EAX.
			gen asm call: gen asm EAX ]

 ! !

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/5/2010 19:19'!
primAllocatePage: bytesToAllocate
	| ptr buf |

	buf := ByteArray new: 4.

	self 
		mmapLength: bytesToAllocate 
		prot: self protFlags 
		flags: self mapFlags 
		into: buf.

	ptr := buf unsignedLongAt: 1 bigEndian: false.
	ptr = MAP_FAILED ifTrue: [ self error: 'Unable to map virtual memory' ].
	
	^ ptr

 ! !

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/5/2010 19:19'!
primFreePage: aMemoryPage

	"call munmap()"

	| res |
	res := self unmap: aMemoryPage address length: aMemoryPage length.
	res = 0 ifFalse: [ self error: 'error during unmapping virtual memory page' ]! !

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/5/2010 18:36'!
protFlags
	"since NativeBoost using heap for placing executable code there,
	it is important that all allocated memory should be executable"
	^  (PROT_READ bitOr: PROT_WRITE) bitOr: PROT_EXEC! !

!NBMacExternalHeapManager methodsFor: 'as yet unclassified' stamp: 'IgorStasenko 11/24/2012 17:22'!
unmap: addr  length: len 
		
	<primitive: #primitiveNativeCall module: #NativeBoostPlugin error: errorCode>

	"unmap memory page"

	^ self nbCallout
		function: #( int munmap (uint addr, ulong  len)  ) 
		module: RTLD_DEFAULT
 ! !

NativeBoostLinux32 subclass: #NativeBoostMac32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'NBMacConstants'
	category: 'NativeBoost-Mac'!

!NativeBoostMac32 class methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 12/4/2010 22:21'!
targetPlatformId
	
	^ Mac32PlatformId ! !

!NativeBoostMac32 methodsFor: 'platform-specific' stamp: 'IgorStasenko 2/2/2012 16:00'!
VMModule
	"answer a C library path, or module handle,
	use in FFI callouts as module: argument to call C run-time library function(s)"
	
	^ self CLibrary! !

!NativeBoostMac32 methodsFor: 'bootstrapping' stamp: 'IgorStasenko 11/24/2012 16:03'!
bootstrapGetGlobalSymbolPointer: symbolName into: returnValueBuffer
	" call  dlsym(RTLD_DEFAULT, symbolName) "
	" override, since mac using different RTLD_DEFAULT constant value "
	
	<primitive: 'primitiveNativeCall' module: 'NativeBoostPlugin' error: errorCode>
	^ self nbCallout
		options: #( - optUseStackPointer);
		function: #( NBBootstrapUlong (RTLD_DEFAULT , String symbolName) )
		emit: [:gen |	 | asm |
				asm := gen asm.
				asm mov: (self primitiveDlsym asUImm) to: asm EAX.
				asm call: asm EAX.
		]! !

!NativeBoostMac32 methodsFor: 'bootstrapping' stamp: 'IgorStasenko 11/24/2012 16:03'!
getEntryPointStackOffset
	" call  dlsym(RTLD_DEFAULT, symbolName) "
	" override, since mac using different RTLD_DEFAULT constant value "
	
	<primitive: 'primitiveNativeCall' module: 'NativeBoostPlugin' error: errorCode>
	^ self nbCallout
		function: #( ulong () )
		emit: [:gen |	 | asm |
				asm := gen asm.
				asm mov: asm ESP to: asm EAX;
					and: asm EAX with: 16r0F.
		].

	! !

!NativeBoostMac32 methodsFor: 'bootstrapping' stamp: 'Igor.Stasenko 12/5/2010 18:32'!
initializeExternalHeap

	"initialize external memory heap"
		
	heap := NBMacExternalHeapManager new.

	! !

!NativeBoostMac32 methodsFor: 'assembler' stamp: 'IgorStasenko 8/5/2011 08:48'!
newAssembler
	^ AJx86Assembler new! !

!NativeBoostMac32 methodsFor: 'platform-specific' stamp: 'IgorStasenko 8/2/2011 06:33'!
stackAlignment

	^ 16! !
NBMacConstants initialize!
